==> dune-fem-2.4.1.tar.gz: dune-fem-2.4.1/dune/fem/space/lagrange/test/generic.cc <==
/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2015 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2015 Andreas Dedner
  Copyright (C) 2005        Adrian Burri
  Copyright (C) 2005 - 2015 Mirko Kraenkel
  Copyright (C) 2006 - 2015 Christoph Gersbacher
  Copyright (C) 2006 - 2015 Martin Nolte
  Copyright (C) 2011 - 2015 Tobias Malkmus
  Copyright (C) 2012 - 2015 Stefan Girke
  Copyright (C) 2013 - 2015 Claus-Justus Heine
  Copyright (C) 2013 - 2014 Janick Gerstenberger
  Copyright (C) 2013        Sven Kaulman
  Copyright (C) 2013        Tom Ranner
  Copyright (C) 2015        Marco Agnese
  Copyright (C) 2015        Martin Alkaemper


  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#include <config.h>

#include <dune/common/forloop.hh>

#include "../shapefunctionset.hh"
#include "../lagrangepoints.hh"

using namespace Dune;
using namespace Fem;

#define VERBOSITY_LEVEL 0

#ifndef DIMENSION
#define DIMENSION 2
#endif

#ifndef POLORDER
#define POLORDER 2
#endif

#ifndef GEOMETRYTYPE
#define GEOMETRYTYPE simplex
#endif

typedef FunctionSpace< double, double, DIMENSION, 1 > FunctionSpaceType;

typedef FunctionSpaceType :: DomainType DomainType;
typedef FunctionSpaceType :: RangeType RangeType;

typedef Fem::LagrangeShapeFunctionFactory< FunctionSpaceType, POLORDER > ShapeFunctionFactoryType;
typedef ShapeFunctionFactoryType::ShapeFunctionType ShapeFunctionType;

typedef Fem::LagrangePoint< GeometryType :: GEOMETRYTYPE, DIMENSION, POLORDER >
  LagrangePointType;
typedef Fem::LagrangePointListImplementation
  < double, GeometryType :: GEOMETRYTYPE, DIMENSION, POLORDER >
  LagrangePointListType;



template< int codim >
struct PrintMaxDofs
{
  static void apply ()
  {
    const unsigned int maxCodimDofs = LagrangePointType::template Codim< codim >::maxDofs();
    std::cout << "MaxDofs< " << codim << " >: " << maxCodimDofs << std::endl;
  }
};



int main( int argc, char **argv )
{
  GeometryType geometryType( GeometryType::GEOMETRYTYPE, DIMENSION );
  ShapeFunctionFactoryType shapeFunctionFactory ( geometryType );

  const unsigned int numShapeFunctions = shapeFunctionFactory.numShapeFunctions();

  std::cout << "Number of shape functions: " << numShapeFunctions;
  std::cout << std::endl << std::endl;

  ForLoop< PrintMaxDofs, 0, DIMENSION >::apply();
  std::cout << std::endl;

  unsigned int errors = 0;
  unsigned int indexErrors = 0;
  unsigned int pointSetErrors = 0;
  LagrangePointListType pointSet( 0 );
  for( unsigned int i = 0; i < numShapeFunctions; ++i )
  {
    LagrangePointType point( i );

    DomainType x;
    point.local( x );
    x -= pointSet.point( i );
    if( x.two_norm() >= 1e-14 )
      ++pointSetErrors;

    unsigned int codim, setCodim;
    unsigned int subEntity, setSubEntity;
    unsigned int dofNumber, setDofNumber;
    point.dofSubEntity( codim, subEntity, dofNumber );
    pointSet.dofSubEntity( i, setCodim, setSubEntity, setDofNumber );
    if( (setCodim != codim) || (setSubEntity != subEntity)
        || (setDofNumber != dofNumber) )
      ++pointSetErrors;

    unsigned int numDofs = point.numDofs( codim, subEntity );
    if( numDofs != pointSet.numDofs( codim, subEntity ) )
      ++pointSetErrors;

    unsigned int entityDofNumber
      = point.entityDofNumber( codim, subEntity, dofNumber );
    if( entityDofNumber != pointSet.entityDofNumber( codim, subEntity, dofNumber ) )
      ++pointSetErrors;

    if( dofNumber >= numDofs )
      ++errors;
    if( entityDofNumber != i )
      ++indexErrors;

    std :: cout << i << ": x = " << pointSet.point( i );
    std :: cout << ", height = " << point.height();
    std :: cout << ", codim = " << codim << ", subEntity = " << subEntity;
    std :: cout << ", dofNumber = " << dofNumber << " ( " << numDofs << " )";
    std :: cout << ", entityDofNumber = " << entityDofNumber;
    std :: cout << std :: endl;
  }
  std :: cout << std :: endl;
  std :: cout << "DoFs with dofNumber >= numDofs: " << errors << std :: endl;
  std :: cout << "DoFs entityDofNumber( dofSubEntity ) != i: " << indexErrors
              << std :: endl;
  std :: cout << "Errors in point set: " << pointSetErrors << std :: endl;
  std :: cout << std :: endl;

  errors = 0;
  for( unsigned int j = 0; j < numShapeFunctions; ++j )
  {
    ShapeFunctionType *shapeFunction = shapeFunctionFactory.createShapeFunction( j );

    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << "Shape function: " << j << std :: endl;
    #endif
    for( unsigned int i = 0; i < numShapeFunctions; ++i )
    {
      LagrangePointType point( i );

      DomainType x;
      point.local( x );

      RangeType phi;
      FieldVector< int, 0 > derivative;
      shapeFunction->evaluate( x, phi );

      double expected = ((i == j) ? 1.0 : 0.0);
      if( fabs( phi[ 0 ] - expected ) > 1e-8 )
        ++errors;

      #if (VERBOSITY_LEVEL >= 1)
        std :: cout << i << ": x = " << x;
        std :: cout << ", f( x ) = " << phi[ 0 ];
        std :: cout << std :: endl;
      #endif
    }

    delete shapeFunction;
    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << std :: endl;
    #endif
  }
  std :: cout << "Shape function evaluation errors: " << errors << std :: endl;

  errors = 0;
  for( unsigned int i = 0; i < numShapeFunctions; ++i )
  {
    LagrangePointType point( i );

    DomainType x;
    point.local( x );

    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << "Lagrange point " << i << ": " << x << std :: endl;
    #endif

    typedef ShapeFunctionType::JacobianRangeType JacobianRangeType;
    JacobianRangeType jacobian( 0 );
    for( unsigned int j = 0; j < numShapeFunctions; ++j )
    {
      ShapeFunctionType *shapeFunction = shapeFunctionFactory.createShapeFunction( j );

      JacobianRangeType tmp;
      shapeFunction->jacobian( x, tmp );
      #if (VERBOSITY_LEVEL >= 1)
        std :: cout << "ShapeFunction " << j << ", jacobian: " << tmp << std :: endl;
      #endif

      jacobian += tmp;

      delete shapeFunction;
    }

    if( fabs( jacobian[ 0 ].two_norm() ) > 1e-8 )
      ++errors;

    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << std :: endl;
    #endif
  }
  std :: cout << "Shape function derivtive summation errors: " << errors << std :: endl;

  return 0;
}

==> dune-fem-2.8.0.dev20201201.tar.gz: dune-fem-2.8.0.dev20201201/dune/fem/space/lagrange/test/generic.cc <==
#include <config.h>

#include <dune/common/hybridutilities.hh>

#include <dune/geometry/type.hh>

#include "../shapefunctionset.hh"
#include "../lagrangepoints.hh"

using namespace Dune;
using namespace Fem;

#define VERBOSITY_LEVEL 0

#ifndef DIMENSION
#define DIMENSION 2
#endif

#ifndef POLORDER
#define POLORDER 2
#endif

#ifndef TOPOLOGYTYPE
#define TOPOLOGYTYPE SimplexTopology
#endif

typedef FunctionSpace< double, double, DIMENSION, 1 > FunctionSpaceType;

typedef FunctionSpaceType :: DomainType DomainType;
typedef FunctionSpaceType :: RangeType RangeType;

typedef Fem::LagrangeShapeFunctionFactory< FunctionSpaceType, POLORDER > ShapeFunctionFactoryType;
typedef ShapeFunctionFactoryType::ShapeFunctionType ShapeFunctionType;

typedef Fem::LagrangePoint< Dune::Impl::TOPOLOGYTYPE<DIMENSION>::type::id, DIMENSION, POLORDER > LagrangePointType;
typedef Fem::LagrangePointListImplementation< double, Dune::Impl::TOPOLOGYTYPE<DIMENSION>::type::id, DIMENSION, POLORDER >
  LagrangePointListType;


int main( int argc, char **argv )
{
  GeometryType geometryType( Dune::Impl::TOPOLOGYTYPE<DIMENSION> ::type::id, DIMENSION );
  ShapeFunctionFactoryType shapeFunctionFactory ( geometryType );

  const unsigned int numShapeFunctions = shapeFunctionFactory.numShapeFunctions();

  std::cout << "Number of shape functions: " << numShapeFunctions;
  std::cout << std::endl << std::endl;

  Hybrid::forEach( std::make_index_sequence< DIMENSION+1 >{},
    [ & ]( auto i ){ std::cout << "MaxDofs< " << i << " >: " << LagrangePointType::template Codim< i >::maxDofs() << std::endl; } );
  std::cout << std::endl;

  unsigned int errors = 0;
  unsigned int indexErrors = 0;
  unsigned int pointSetErrors = 0;
  LagrangePointListType pointSet( 0 );
  for( unsigned int i = 0; i < numShapeFunctions; ++i )
  {
    LagrangePointType point( i );

    DomainType x;
    point.local( x );
    x -= pointSet.point( i );
    if( x.two_norm() >= 1e-14 )
      ++pointSetErrors;

    unsigned int codim, setCodim;
    unsigned int subEntity, setSubEntity;
    unsigned int dofNumber, setDofNumber;
    point.dofSubEntity( codim, subEntity, dofNumber );
    pointSet.dofSubEntity( i, setCodim, setSubEntity, setDofNumber );
    if( (setCodim != codim) || (setSubEntity != subEntity)
        || (setDofNumber != dofNumber) )
      ++pointSetErrors;

    unsigned int numDofs = point.numDofs( codim, subEntity );
    if( numDofs != pointSet.numDofs( codim, subEntity ) )
      ++pointSetErrors;

    unsigned int entityDofNumber
      = point.entityDofNumber( codim, subEntity, dofNumber );
    if( entityDofNumber != pointSet.entityDofNumber( codim, subEntity, dofNumber ) )
      ++pointSetErrors;

    if( dofNumber >= numDofs )
      ++errors;
    if( entityDofNumber != i )
      ++indexErrors;

    std :: cout << i << ": x = " << pointSet.point( i );
    std :: cout << ", height = " << point.height();
    std :: cout << ", codim = " << codim << ", subEntity = " << subEntity;
    std :: cout << ", dofNumber = " << dofNumber << " ( " << numDofs << " )";
    std :: cout << ", entityDofNumber = " << entityDofNumber;
    std :: cout << std :: endl;
  }
  std :: cout << std :: endl;
  std :: cout << "DoFs with dofNumber >= numDofs: " << errors << std :: endl;
  std :: cout << "DoFs entityDofNumber( dofSubEntity ) != i: " << indexErrors
              << std :: endl;
  std :: cout << "Errors in point set: " << pointSetErrors << std :: endl;
  std :: cout << std :: endl;

  errors = 0;
  for( unsigned int j = 0; j < numShapeFunctions; ++j )
  {
    ShapeFunctionType *shapeFunction = shapeFunctionFactory.createShapeFunction( j );

    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << "Shape function: " << j << std :: endl;
    #endif
    for( unsigned int i = 0; i < numShapeFunctions; ++i )
    {
      LagrangePointType point( i );

      DomainType x;
      point.local( x );

      RangeType phi;
      shapeFunction->evaluate( x, phi );

      double expected = ((i == j) ? 1.0 : 0.0);
      if( fabs( phi[ 0 ] - expected ) > 1e-8 )
        ++errors;

      #if (VERBOSITY_LEVEL >= 1)
        std :: cout << i << ": x = " << x;
        std :: cout << ", f( x ) = " << phi[ 0 ];
        std :: cout << std :: endl;
      #endif
    }

    delete shapeFunction;
    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << std :: endl;
    #endif
  }
  std :: cout << "Shape function evaluation errors: " << errors << std :: endl;

  errors = 0;
  for( unsigned int i = 0; i < numShapeFunctions; ++i )
  {
    LagrangePointType point( i );

    DomainType x;
    point.local( x );

    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << "Lagrange point " << i << ": " << x << std :: endl;
    #endif

    typedef ShapeFunctionType::JacobianRangeType JacobianRangeType;
    JacobianRangeType jacobian( 0 );
    for( unsigned int j = 0; j < numShapeFunctions; ++j )
    {
      ShapeFunctionType *shapeFunction = shapeFunctionFactory.createShapeFunction( j );

      JacobianRangeType tmp;
      shapeFunction->jacobian( x, tmp );
      #if (VERBOSITY_LEVEL >= 1)
        std :: cout << "ShapeFunction " << j << ", jacobian: " << tmp << std :: endl;
      #endif

      jacobian += tmp;

      delete shapeFunction;
    }

    if( fabs( jacobian[ 0 ].two_norm() ) > 1e-8 )
      ++errors;

    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << std :: endl;
    #endif
  }
  std :: cout << "Shape function derivtive summation errors: " << errors << std :: endl;

  return 0;
}
